import { defaultKeymap, indentWithTab } from '@codemirror/commands';
import { indentUnit } from '@codemirror/language';
import { Compartment, EditorState } from '@codemirror/state';
import { EditorView, keymap } from '@codemirror/view';
import { map } from 'nanostores';

// src/index.ts
var withCodemirrorInstance = () => map({
  view: null,
  extensions: null,
  value: null
});
var codemirror = (node, options) => {
  if (!options)
    throw new Error("No options provided. At least `value` is required.");
  let {
    value,
    instanceStore,
    diagnostics,
    onChangeBehavior = { kind: "debounce", duration: 50 }
  } = options;
  let fulfill_editor_initialized;
  let editor_initialized = new Promise((r) => fulfill_editor_initialized = r);
  let view;
  let internal_extensions = [];
  const setup_compartment = new Compartment();
  const lang_compartment = new Compartment();
  const theming_compartment = new Compartment();
  const tabs_compartment = new Compartment();
  const readonly_compartment = new Compartment();
  const extensions_compartment = new Compartment();
  const autocomplete_compartment = new Compartment();
  async function make_extensions(options2) {
    return [
      // User extensions matter the most, keep em on top
      extensions_compartment.of(get_user_extensions(options2)),
      // Autocomplete may come built in with setup: basic, so always keep it above setup_compartment
      autocomplete_compartment.of(await get_autocompletion(options2)),
      setup_compartment.of(await get_setup(options2) ?? []),
      // Needs to be under `setup` because setup, if there, will add the indentWithTab
      keymap.of([...defaultKeymap, ...options2.useTabs ? [indentWithTab] : []]),
      lang_compartment.of(await get_lang(options2)),
      theming_compartment.of(get_theme(options2)),
      tabs_compartment.of(await get_tab_setting(options2)),
      readonly_compartment.of(get_readonly(options2))
    ];
  }
  function handle_change(tr) {
    const new_value = view.state.doc.toString();
    if (new_value === value)
      return;
    if (new_value !== value) {
      value = new_value;
      node.dispatchEvent(new CustomEvent("codemirror:textChange", { detail: value }));
      options.onTextChange?.(value);
    }
    instanceStore?.set({ value, view, extensions: internal_extensions });
    node.dispatchEvent(new CustomEvent("codemirror:change", { detail: tr }));
    options.onChange?.(tr);
  }
  const { kind: behaviorKind = "debounce", duration: behaviorDuration = 50 } = onChangeBehavior;
  let on_change = behaviorKind === "debounce" ? debounce(handle_change, behaviorDuration) : throttle(handle_change, behaviorDuration);
  (async () => {
    internal_extensions = await make_extensions(options);
    const state = EditorState.create({
      doc: value,
      extensions: internal_extensions,
      selection: {
        anchor: options.cursorPos ?? 0,
        head: options.cursorPos ?? 0
      }
    });
    view = new EditorView({
      state,
      parent: node,
      dispatch(tr) {
        view.update([tr]);
        on_change(tr);
      }
    });
    make_diagnostics(view, diagnostics);
    instanceStore?.set({
      view,
      extensions: internal_extensions,
      value
    });
    fulfill_editor_initialized();
  })();
  return {
    async update(new_options) {
      await editor_initialized;
      const transaction = {};
      if (value !== new_options.value) {
        value = new_options.value;
        transaction.changes = {
          from: 0,
          to: view.state.doc.length,
          insert: value
        };
      }
      if (typeof new_options.cursorPos !== "undefined" && options.cursorPos !== new_options.cursorPos) {
        transaction.selection = {
          anchor: new_options.cursorPos ?? 0,
          head: new_options.cursorPos ?? 0
        };
      }
      async function append_effect(compartment, options_list, factory) {
        transaction.effects = transaction.effects ?? [];
        for (const option_name of options_list) {
          const new_option = new_options[option_name];
          const old_option = options[option_name];
          const effects = transaction.effects;
          if (typeof new_option !== "undefined") {
            if (new_option !== old_option) {
              effects.push(compartment.reconfigure(await factory(new_options)));
            }
          } else {
            return effects.push(compartment.reconfigure([]));
          }
        }
      }
      await Promise.all([
        append_effect(setup_compartment, ["setup"], get_setup),
        append_effect(lang_compartment, ["lang"], get_lang),
        append_effect(tabs_compartment, ["useTabs", "tabSize"], get_tab_setting),
        append_effect(theming_compartment, ["theme"], get_theme),
        append_effect(extensions_compartment, ["extensions"], get_user_extensions),
        append_effect(readonly_compartment, ["readonly"], get_readonly),
        append_effect(autocomplete_compartment, ["autocomplete"], get_autocompletion)
      ]);
      view.dispatch(transaction);
      make_diagnostics(view, new_options.diagnostics);
      internal_extensions = await make_extensions(new_options);
      instanceStore?.set({
        view,
        extensions: internal_extensions,
        value
      });
      const { kind: behaviorKind2 = "debounce", duration: behaviorDuration2 = 50 } = new_options.onChangeBehavior ?? { kind: "debounce", duration: 50 };
      if (options.onChangeBehavior?.kind !== behaviorKind2 || options.onChangeBehavior.duration !== behaviorDuration2) {
        on_change = behaviorKind2 === "debounce" ? debounce(handle_change, behaviorDuration2) : throttle(handle_change, behaviorDuration2);
      }
      options = new_options;
    },
    destroy() {
      editor_initialized.then(() => view?.destroy());
    }
  };
};
async function get_setup(options) {
  const { setup } = options;
  if (!setup)
    return [];
  if (setup === "basic")
    return (await import('./basic-setup-BJSL5TOQ.js')).default(options);
  if (setup === "minimal")
    return (await import('./minimal-setup-Y523JPZT.js')).default(options);
  throw new Error(
    "`setup` can only be `basic` or `minimal`. If you wish to provide another setup, pass through `extensions` prop."
  );
}
async function get_lang({ lang, langMap }) {
  if (!lang)
    return [];
  if (typeof lang === "string") {
    if (!langMap)
      throw new Error("`langMap` is required when `lang` is a string.");
    if (!(lang in langMap))
      throw new Error(`Language "${lang}" is not defined in \`langMap\`.`);
    const lang_support = await langMap[lang]();
    return lang_support;
  }
  return lang;
}
function get_theme({ theme, styles }) {
  return [theme, styles && EditorView.theme(styles)].filter(Boolean);
}
async function get_tab_setting({ useTabs = false, tabSize = 2 }) {
  return [EditorState.tabSize.of(tabSize), indentUnit.of(useTabs ? "	" : " ".repeat(tabSize))];
}
async function get_autocompletion({ autocomplete }) {
  if (!autocomplete)
    return [];
  const { autocompletion } = await import('@codemirror/autocomplete');
  return autocompletion(typeof autocomplete === "object" && autocomplete ? autocomplete : {});
}
function get_readonly({ readonly }) {
  return EditorState.readOnly.of(!!readonly);
}
function get_user_extensions({ extensions }) {
  return extensions ?? [];
}
async function make_diagnostics(view, diagnostics) {
  if (!diagnostics)
    return;
  const { setDiagnostics } = await import('@codemirror/lint');
  const tr = setDiagnostics(view.state, diagnostics ?? []);
  view.dispatch(tr);
}
function debounce(func, threshold, execAsap = false) {
  let timeout;
  return function debounced(...args) {
    const self = this;
    if (timeout)
      clearTimeout(timeout);
    else if (execAsap)
      func.apply(self, args);
    timeout = setTimeout(delayed, threshold || 100);
    function delayed() {
      if (!execAsap)
        func.apply(self, args);
      timeout = null;
    }
  };
}
function throttle(func, threshold) {
  let last_args;
  let should_wait = false;
  function timeout_function(self) {
    if (last_args) {
      func.apply(self, last_args);
      setTimeout(timeout_function, threshold, self);
      last_args = null;
      return;
    }
    should_wait = false;
  }
  return function throttled(...args) {
    const self = this;
    if (should_wait) {
      last_args = args;
      return;
    }
    func.apply(self, args);
    should_wait = true;
    setTimeout(timeout_function, threshold, self);
  };
}

export { codemirror, withCodemirrorInstance };
